""" Functions for code quality control of Jupyter Notebooks. """
#pylint: disable=import-outside-toplevel
import os

from .core import StringWithDisabledRepr, get_notebook_path, notebook_to_script



PYLINTRC_TEMPLATE = """
[DESIGN]
max-line-length={max_line_length}
variable-rgx=(.*[a-z][a-z0-9_]{{1,30}}|[a-z_])$ # snake_case + single letters
argument-rgx=(.*[a-z][a-z0-9_]{{1,30}}|[a-z_])$ # snake_case + single letters

[MESSAGE CONTROL]
disable=too-few-public-methods, unsubscriptable-object, no-member, too-many-lines, too-many-locals,
        arguments-renamed, arguments-differ, multiple-statements, invalid-name,
        missing-module-docstring, missing-final-newline, redefined-outer-name,
        wrong-import-position, no-name-in-module, import-error, unused-wildcard-import,
        unnecessary-semicolon, trailing-whitespace, trailing-newlines,
        {disable}

good-names=bar, df, fn
additional-builtins=display, get_ipython
"""


def generate_pylintrc(path, disable=(), enable=(), max_line_length=120, **pylint_params):
    """ Create `pylintrc` file.

    Parameters
    ----------
    path : str
        Path to save the file.
    disable : sequence
        Which checks to disable. Each element should be either a code or a name of the check.
    enable : sequence
        Which checks to enable. Each element should be either a code or a name of the check.
        Has priority over `disable`.
    max_line_length : int
        Allowed line length.
    pylint_params : dict
        Additional parameter of linting. Each is converted to a separate valid entry in the `pylintrc` file.
    """
    disable = [disable] if isinstance(disable, str) else disable
    enable = [enable] if isinstance(enable, str) else enable
    disable = [item.replace('_', '-') for item in disable]
    enable = [item.replace('_', '-') for item in enable]

    pylintrc = PYLINTRC_TEMPLATE.format(disable=','.join(disable),
                                        max_line_length=max_line_length)

    for item in enable:
        if item in pylintrc:
            pylintrc = (pylintrc.replace(item, '')
                                .replace('=, ', '=')
                                .replace(', ,', ','))

    for key, value in pylint_params.items():
        key = key.replace('_', '-')
        if isinstance(value, list):
            value = ', '.join(value)
        pylintrc += f'\n{key}={value}'

    with open(path, 'w', encoding='utf-8') as file:
        file.write(pylintrc)

    return pylintrc


def pylint_notebook(path=None, options=(), disable=(), enable=(), printer=print,
                    remove_files=True, return_info=False, **pylint_params):
    """ Execute `pylint` for a provided Jupyter Notebook.

    Under the hood, roughly does the following:
        - Creates a `.pylintrc` file next to the `path`.
        - Converts the notebook to `.py` file next to the `path`.
        - Runs `pylint` with additional options.
        - Create a report and display it, if needed.

    Parameters
    ----------
    path : str, optional
        Path to the Jupyter notebook. If not provided, the current notebook is used.
    options : sequence
        Additional options for `pylint` execution.
    printer : callable or None
        Function to display the report.
    remove_files : bool
        Whether to remove `.pylintrc` and `.py` files after the execution.
    return_info : bool
        Whether to return a dictionary with intermediate results.
        It contains the notebook code string, as well as `pylint` stdout and stderr.
    disable : sequence
        Which checks to disable. Each element should be either a code or a name of the check.
    enable : sequence
        Which checks to enable. Each element should be either a code or a name of the check.
        Has priority over `disable`.
    max_line_length : int
        Allowed line length.
    pylint_params : dict
        Additional parameter of linting. Each is converted to a separate valid entry in the `pylintrc` file.
    """
    from pylint import epylint as lint
    path = path or get_notebook_path()
    if path is None:
        raise ValueError('Provide path to Jupyter Notebook or run `pylint_notebook` inside of it!')

    # Convert notebook to a script
    path_script = os.path.splitext(path)[0] + '.py'
    script_name = os.path.basename(path_script)

    code, cell_line_numbers = notebook_to_script(path_notebook=path, path_script=path_script, return_info=True).values()

    # Create pylintrc file
    path_pylintrc = os.path.splitext(path)[0] + '.pylintrc'
    pylintrc = generate_pylintrc(path_pylintrc, disable=disable, enable=enable, **pylint_params)

    # Run pylint on script with pylintrc configuration
    pylint_cmdline = ' '.join([path_pylintrc, f'--rcfile {path_pylintrc}', *options])
    pylint_stdout, pylint_stderr = lint.py_run(pylint_cmdline, return_std=True)

    errors = pylint_stderr.getvalue()
    report = pylint_stdout.getvalue()

    # Prepare custom report
    output = []

    for line in report.split('\n'):
        if 'rated' in line:
            output.insert(0, line.strip(' '))
            output.insert(1, '–' * (len(line) - 1))

        elif path_script in line or script_name in line:
            line = line.replace(path_script, '__').replace(script_name, '__')

            # Locate the cell and line inside the cell
            code_line_number = int(line.split(':')[1])
            for cell_number, cell_ranges in cell_line_numbers.items():
                if code_line_number in cell_ranges:
                    cell_line_number = code_line_number - cell_ranges[0]
                    break
            else:
                cell_number, cell_line_number = -1, code_line_number

            # Find error_code and error_name: for example, `C0123` and `invalid-name`
            position_left  = line.find('(') + 1
            position_right = line.find(')') - 1
            error_message = line[position_left : position_right]
            error_human_message = line[position_right + 2:]
            error_code, error_name, *_ = error_message.split(',')
            error_name = error_name.strip()

            # Make new message
            message = f'Cell {cell_number}:{cell_line_number}, code={error_code}, name={error_name}'
            message += f'\n     Pylint message ::: {error_human_message}\n'
            output.append(message)

    output = '\n'.join(output).strip()

    if remove_files:
        os.remove(path_script)
        os.remove(path_pylintrc)

    if printer is not None:
        printer(output)

    if return_info:
        enumerated_code = code.split('\n')
        n_digits = len(str(len(enumerated_code)))
        enumerated_code = [f'{i:0>{n_digits}}    ' + item
                           for i, item in enumerate(enumerated_code, start=1)]
        enumerated_code = '\n'.join(enumerated_code)

        return {
            'report' : StringWithDisabledRepr(output),

            'code' : StringWithDisabledRepr(code),
            'enumerated_code' : StringWithDisabledRepr(enumerated_code),

            'pylintrc' : StringWithDisabledRepr(pylintrc),
            'pylint_errors' : StringWithDisabledRepr(errors),
            'pylint_report' : StringWithDisabledRepr(report),
        }
    return None



""" ResourceEntry -- a dict-like class to hold all properties (Resources) of an entry. """
from datetime import datetime

from .resource import Resource
from .utils import format_memory

class ResourceEntry(dict):
    """ Dictionary to hold all properties (Resources) of an entry.
    An entry is any monitored entity -- Python process, Jupyter Notebook or device.
    For more about Resources, refer to its class documentation.

    `getitem` is overloaded to recognize Resource aliases (like `memory_util`) as actual Resource enumeration members.

    The main method of this class, `format`, is used to create a string representation of a single requested property
    from the contained information. As some of the requested columns require multiple values from the dictionary,
    we can't use individual key-value pairs to create string representation on their own.
    For example, the `DEVICE_MEMORY` column show the `'used_memory / total_memory MB'` information and
    requires multiple items from the ResourceEntry at the same time.
    """
    def __getitem__(self, key):
        key = Resource.parse_alias(key)
        return super().__getitem__(key)

    def get(self, key, default=None):
        key = Resource.parse_alias(key)
        return super().get(key, default)

    def to_format_data(self, resource, terminal, **kwargs):
        """ Create a string template and data for a given `resource`. Essentially, a huge switch on `resource` type.
        For more information about formatting refer to `ResourceTable.format` method.

        Parameters
        ----------
        resource : member of Resource
            Column to create string representation for.
        terminal : blessed.Terminal
            Terminal to use for text formatting and color control sequences.
        kwargs : dict
            Other parameters for string creation like memory format, width, etc.
        """
        #pylint: disable=too-many-statements
        resource = Resource.parse_alias(resource)
        default_string = '-'
        style, string = None, None
        data = self.get(resource, None)

        # Process description
        if resource in [Resource.PATH, Resource.CMDLINE, Resource.STATUS,
                        Resource.PID, Resource.PPID, Resource.NGID, Resource.HOST_PID, Resource.PYTHON_PPID,
                        Resource.DEVICE_PROCESS_PID,]:
            pass

        elif resource in [Resource.NAME, Resource.TYPE]:
            if data is not None:
                if 'zombie' in data or 'container' in data:
                    style = terminal.red
                if '/' in data:
                    data = '~' + data.split('/')[-1]

        elif resource == Resource.CREATE_TIME:
            if data is not None:
                data = datetime.fromtimestamp(data).strftime("%Y-%m-%d %H:%M:%S")
        elif resource == Resource.KERNEL:
            data = data.split('-')[0] if data is not None else 'N/A'

        # Process resources
        elif resource == Resource.CPU:
            process = self[Resource.PROCESS]
            if process is not None:
                data = process.cpu_percent()
                data = round(data)

                style = terminal.bold if data > 30 else ''
                string = f'{data}%' # don't use the `％` symbol as it is not unit wide

        elif resource == Resource.RSS:
            if data is not None:
                rounded, unit = format_memory(data, format=kwargs['process_memory_format'])
                data = f'{rounded} {unit}'

        # Device description
        elif resource == Resource.DEVICE_ID:
            if data is not None:
                device_name = (self[Resource.DEVICE_NAME].replace('NVIDIA', '').replace('RTX', '')
                               .replace('  ', ' ').strip())
                string = f'{device_name} {terminal.cyan}[{data}]'

        elif resource == Resource.DEVICE_SHORT_ID:
            data = self.get(Resource.DEVICE_ID, None)
            if data is not None:
                data = data if data is not None else default_string
                device_name = (self[Resource.DEVICE_NAME]
                               .replace('NVIDIA', '').replace('RTX', '').replace('GeForce', '')
                               .replace('  ', ' ').strip())
                string = f'{device_name} [{data}]'

        # Device resources
        elif resource == Resource.DEVICE_MEMORY_USED:
            if data is not None:
                style = terminal.bold if data > 10*1024*1024 else ''

                memory_format = kwargs['device_memory_format']
                used, unit = format_memory(data, format=memory_format)
                total, unit = format_memory(self[Resource.DEVICE_MEMORY_TOTAL], format=memory_format)

                n_digits = len(str(total))
                string = (f'{terminal.normal + terminal.yellow}{style}{used:>{n_digits}}'
                          f'{terminal.normal + terminal.bold} / '
                          f'{terminal.normal + terminal.yellow}{style}{total} '
                          f'{terminal.normal + terminal.bold}{unit}')

        elif resource == Resource.DEVICE_PROCESS_MEMORY_USED:
            if data is not None:
                style = terminal.bold if data > 10*1024*1024 else ''

                memory_format = kwargs['device_memory_format']
                used_process, unit = format_memory(data, format=memory_format)
                used_device, unit = format_memory(self[Resource.DEVICE_MEMORY_USED], format=memory_format)
                total, _ = format_memory(self[Resource.DEVICE_MEMORY_TOTAL], format=memory_format)

                n_digits = len(str(total))
                string = (f'{terminal.normal + terminal.yellow}{style}{used_process:>{n_digits}}'
                          f'{terminal.normal + terminal.bold} / '
                          f'{terminal.normal + terminal.yellow}{style}{max(used_device, used_process):>{n_digits}}'
                          f'{terminal.normal + terminal.bold} / '
                          f'{terminal.normal + terminal.yellow}{style}{total} '
                          f'{terminal.normal + terminal.bold}{unit}')
            else:
                # Fallback to total device memory usage, if possible
                if self.get(Resource.DEVICE_MEMORY_USED, None) is not None:
                    entry = {key : self[key] for key in [Resource.DEVICE_MEMORY_USED, Resource.DEVICE_MEMORY_TOTAL]}
                    entry = ResourceEntry(entry)
                    style, string = entry.to_format_data(resource=Resource.DEVICE_MEMORY_USED,
                                                         terminal=terminal, **kwargs)

        elif resource == Resource.DEVICE_POWER_USED:
            if data is not None:
                power_used = data // 1000
                power_total = self[Resource.DEVICE_POWER_TOTAL] // 1000
                string = f'{power_used:>3}/{power_total:>3} W'

        elif resource in [Resource.DEVICE_FAN, Resource.DEVICE_UTIL, Resource.DEVICE_UTIL_MA]:
            if data is not None:
                style = terminal.bold if data >= 30 else ''
                string = f'{data}%' # don't use the `％` symbol as it is not unit wide

                if kwargs.get('bar'):
                    if data < 30:
                        bar_color = terminal.on_red
                    elif data < 70:
                        bar_color = terminal.on_yellow
                    else:
                        bar_color = terminal.on_green

                    string = f'{string:^4}'.center(9)

                    split = data // 10
                    string_1 = f'{terminal.normal}{bar_color}{style}{string[:split]}'
                    string_2 = f'{terminal.normal}{style}{string[split:]}' # can add {terminal.on_white}
                    string = string_1 + string_2

        elif resource == Resource.DEVICE_TEMP:
            if data is not None:
                style = terminal.bold if data >= 40 else ''
                string = f'{data}°C' # don't use the `℃` symbol as it is not unit wide

        # Table delimiters
        elif resource == Resource.TABLE_DELIMITER1:
            string = '┃'
        elif resource == Resource.TABLE_DELIMITER2:
            string = '║'

        # Default values
        if style is None:
            style = ''
        if string is None:
            if data is not None:
                string = str(data)
            else:
                string = default_string
        return style, string



""" !!. """
#pylint: disable=import-outside-toplevel
import os
import re
import json



class StringWithDisabledRepr(str):
    """ String with disabled repr. Used to avoid cluttering repr from function outputs. """
    def __repr__(self):
        """ !!. """
        return f'<StringWithDisabledRepr at {hex(id(self))}. Use `str`/`print` explicitly!>'



def in_notebook():
    """ Return True if in Jupyter notebook and False otherwise. """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        if shell == 'TerminalInteractiveShell':
            return False
        return False
    except NameError:
        return False

def get_notebook_path():
    """ Return the full absolute path of the current Jupyter notebook,
    for example, `/path/path/path/My_notebook_title.ipynb`.

    If run outside Jupyter notebook, returns None.
    """
    if not in_notebook():
        return None

    import requests
    import ipykernel

    # Id of the current running kernel: a string uid
    kernel_id = re.search('kernel-(.*).json', ipykernel.connect.get_connection_file()).group(1)

    # Get running servers for both JupyterLab v2.# and v3.#
    from notebook.notebookapp import list_running_servers as list_running_servers_v2
    from jupyter_server.serverapp import list_running_servers as list_running_servers_v3
    servers = list(list_running_servers_v2()) + list(list_running_servers_v3())

    for server in servers:
        root_dir = server.get('root_dir') or server.get('notebook_dir')
        response = requests.get(requests.compat.urljoin(server['url'], 'api/sessions'),
                                params={'token': server.get('token', '')})

        for params in json.loads(response.text):
            if params['kernel']['id'] == kernel_id:
                relative_path = params['notebook']['path']
                return os.path.join(root_dir, relative_path)
    raise ValueError(f'Unable to find kernel `{kernel_id}` in {len(servers)} servers!')

def get_notebook_name():
    """ Return the title of the current Jupyter notebook without base directory and extension,
    for example, `My_notebook_title`.

    If run outside Jupyter notebook, returns None.
    """
    if not in_notebook():
        return None

    return os.path.splitext(get_notebook_path())[0].split('/')[-1]



def notebook_to_script(path_script, path_notebook=None, ignore_markdown=True, return_info=False):
    """ !!. """
    import nbformat #pylint: disable=import-outside-toplevel
    path_notebook = path_notebook or get_notebook_path()
    if path_notebook is None:
        raise ValueError('Provide path to Jupyter Notebook or run `notebook_to_script` inside of it!')

    # Read notebook as list of cells
    notebook = nbformat.read(path_notebook, as_version=4)

    code_lines = []
    cell_number = 1
    cell_line_numbers = {}

    for cell in notebook['cells']:
        if ignore_markdown and cell['cell_type'] != 'code':
            continue

        cell_lines = cell['source'].split('\n')
        cell_lines.insert(0, f'\n### [{cell_number}] cell')

        # Comment cell/line magics
        for j, line in enumerate(cell_lines):
            if line.startswith('%') or line.startswith('!'):
                cell_lines[j] = '### ' + line

        code_line_number = len(code_lines) + 1
        cell_line_numbers[cell_number] = range(code_line_number, code_line_number + len(cell_lines))

        code_lines.extend([line.strip('\n') for line in cell_lines])
        code_lines.append('')
        cell_number += 1

    code = '\n'.join(code_lines).strip()

    with open(path_script, 'w', encoding='utf-8') as file:
        file.write(code)

    if return_info:
        return {'code': StringWithDisabledRepr(code),
                'cell_line_numbers': cell_line_numbers}
    return None



""" NBstat module: monitoring running Python processes and Notebooks. """
#pylint: disable=wildcard-import
from .resource import Resource
from .resource_entry import ResourceEntry
from .resource_table import ResourceTable
from .resource_inspector import ResourceInspector
from .resource_formatter import ResourceFormatter, NBSTAT_FORMATTER, GPUSTAT_FORMATTER, DEVICESTAT_FORMATTER
from .utils import *

from .cli import *



""" Command line interface of `nbstat`.
Also provides `nbwatch`, `devicestat` and `devicewatch` functions.
"""
#pylint: disable=redefined-outer-name, too-many-nested-blocks
import sys
import traceback
from inspect import cleandoc
from time import time
from argparse import ArgumentParser, RawTextHelpFormatter

from blessed import Terminal

from .resource import Resource
from .resource_formatter import NBSTAT_FORMATTER, DEVICESTAT_FORMATTER, GPUSTAT_FORMATTER
from .resource_inspector import ResourceInspector
from .utils import true_len



def main(name, interval=None):
    """ Run command `name`. If `interval` is given, continuously output it to a terminal in fullscreen mode. """
    # Attach SIGPIPE handler to properly handle broken pipe
    try: # sigpipe not available under windows. just ignore in this case
        import signal # pylint: disable=import-outside-toplevel
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except Exception: # pylint: disable=broad-except
        pass

    # Make parameters for requested view
    inspector = ResourceInspector()
    formatter, view_args = make_parameters(name)

    # Parse interval
    interval = view_args.get('interval') or interval
    interval = max(interval, 0.1) if interval is not None else None
    view_args['interval'] = interval

    # Print table
    if not interval:
        output_once(inspector, name, formatter, view_args)
    else:
        # If in `watch` mode, prepare other handler as well
        other_name = 'devicewatch' if name.startswith('nb') else 'nbwatch'
        other_formatter, other_view_args = make_parameters(other_name)
        other_view_args['interval'] = interval

        output_looped(inspector, name, formatter, view_args,
                      other_name, other_formatter, other_view_args,
                      interval=interval)


def output_once(inspector, name, formatter, view_args):
    """ Output visualization to a stdout once. """
    try:
        view = inspector.get_view(name=name, formatter=formatter, **view_args)
        print(view)
    except Exception as e: # pylint: disable=broad-except
        _ = e
        print('Error on getting system information!' + str(e))
        raise e

def output_looped(inspector, name, formatter, view_args,
                  other_name, other_formatter, other_view_args, interval=0.5):
    """ Output visualization to a stdout once each `interval` seconds in a fullscreen mode. """
    terminal = Terminal()

    initial_view_args = dict(view_args)
    initial_other_view_args = dict(other_view_args)

    with terminal.cbreak(), terminal.fullscreen():
        try: # catches keyboard interrupts to exit gracefully
            counter = 0
            prev_len = 0

            while True:
                counter += 1

                try: # catches `nbstat` exceptions
                    # Get the view
                    start_time = time()
                    view = inspector.get_view(name=name, formatter=formatter, **view_args)
                    current_len = true_len(view)

                    # Select starting position: if needed, redraw the entire screen, otherwise just move cursor position
                    if abs(current_len - prev_len) > 100:
                        start_position = terminal.clear
                    else:
                        start_position = terminal.move(0, 0)
                    prev_len = current_len

                    # Actual print
                    print(start_position, view, ' ', terminal.clear_eol, sep='')

                    # Wait for the input key
                    remaining_time = interval - (time() - start_time)
                    if remaining_time > 0.0:
                        inkey = terminal.inkey(timeout=remaining_time)
                    else:
                        inkey = terminal.inkey(timeout=interval)

                    if inkey:
                        recognized = True

                        # Tab to switch views. Re-print immediately
                        if inkey.code in [terminal.KEY_TAB, terminal.KEY_BTAB]:
                            # Swap every variable
                            name, other_name = other_name, name
                            formatter, other_formatter = other_formatter, formatter
                            view_args, other_view_args = other_view_args, view_args
                            initial_view_args, initial_other_view_args = initial_other_view_args, initial_view_args

                        elif inkey == 's':
                            view_args['separate_index'] = not view_args['separate_index']
                        elif inkey == 'r':
                            view_args = dict(initial_view_args)
                        elif inkey == 'b':
                            formatter.toggle_bars()
                        elif inkey == 'm':
                            if formatter[Resource.DEVICE_UTIL]:
                                formatter[Resource.DEVICE_UTIL_MA] = not formatter[Resource.DEVICE_UTIL_MA]
                        elif inkey == 'q':
                            raise KeyboardInterrupt
                        else:
                            recognized = False

                        if recognized:
                            view = inspector.get_view(name=name, formatter=formatter, **view_args)
                            print(terminal.clear, view, ' ', terminal.clear_eol, sep='')
                        else:
                            print(f'Unrecognized key={inkey}, code={inkey.code}.')

                except Exception: # pylint: disable=broad-except
                    sys.stderr.write(traceback.format_exc())
                    sys.stderr.write('Error on getting system information!')
                    sys.exit(1)

        except KeyboardInterrupt:
            return 0

def nbstat():
    """ Collect all running Jupyter Notebooks and Python scripts. For each of their processes,
    show detailed information on their device usage and other resource utilization.
    """
    return main('nbstat')

def nbwatch():
    """ Collect all running Jupyter Notebooks and Python scripts. For each of their processes,
    continuously show detailed information on their device usage and other resource utilization.
    """
    return main('nbwatch', interval=1.)

def devicestat():
    """ Collect information about all NVIDIA devices. For each of them,
    show detailed information about utilization and processes.
    """
    return main('devicestat')

def devicewatch():
    """ Collect information about all NVIDIA devices. For each of them,
    continuously show detailed information about utilization and processes.
    """
    return main('devicewatch', interval=1.)



NAME_TO_VIEW = {
    'nbstat' : 'nbstat', 'nbwatch' : 'nbstat',
    'devicestat' : 'devicestat', 'devicewatch' : 'devicestat',
    'gpustat' : 'gpustat', 'gpuwatch' : 'gpustat',
}

VIEW_TO_FORMATTER = {
    'nbstat' : NBSTAT_FORMATTER,
    'devicestat' : DEVICESTAT_FORMATTER,
    'gpustat' : GPUSTAT_FORMATTER,
}

DEFAULTS = {
    'index_condition' : None,
    'interval' : 0.0,
    'window': 20,
    'verbose' : 0,

    'show_all' : False,
    'show' : [],
    'hide' : [],
    'hide_similar' : True,

    'add_supheader' : True,
    'add_header' : True,
    'add_footnote' : False,

    'separate_supheader' : False,
    'separate_header' : True,
    'separate_index' : True,

    'force_styling' : True,
    'process_memory_format' : 'GB',
    'device_memory_format' : 'MB',
}

def make_parameters(name):
    """ Parse parameters from command line into dictionary.
    Use `store_const` instad of `store_true` to keep `None` values, if not passed explicitly
    """
    # Set defaults
    defaults = dict(DEFAULTS)
    if 'device' in name:
        defaults.update({'separate_index' : False,})

    if 'watch' in name:
        defaults.update({'add_footnote' : True})

    # Fetch formatter: used to tell which columns can be shown/hidden from the table in documentation
    view = NAME_TO_VIEW[name]
    formatter = VIEW_TO_FORMATTER[view]

    # Command line arguments
    argv = sys.argv[1:]
    if 'device' in name:
        argv = [arg for arg in argv if arg not in ['-v', '-V']]


    # General doc
    docstring = cleandoc(globals()[name].__doc__)
    help_verbose_0 = 'Default `verbose=0` shows only script/notebooks that use devices.' if 'nb' in name else ''
    help_watch = '\nSet `interval` to continuously update displayed table.' if 'watch' not in name else '\n'
    help_keystrokes = (
        'While in the `watch` mode, you can use keystrokes to modify displayed view:'
        '\n  - `tab` — swaps views, from `nbwatch` to `devicewatch` and back.'
        '\n  - `b` — toggles bar representation for some of the resources: in addition to its value, show colored bar.'
        '\n  - `m` — toggles moving averages for some of the resources: values are averaged across the last iterations.'
        '\n  - `s` — toggles table separators.')
    parser = ArgumentParser(description='\n'.join([docstring, help_verbose_0, help_watch, help_keystrokes]),
                            formatter_class=RawTextHelpFormatter)
    linesep = '\n '

    # Positional argument: filtering condition on index
    parser.add_argument('index_condition', nargs='?',
                        help=('Regular expression for filtering entries in the table index. '
                              'For example, `.*.ipynb` allows to look only at Jupyter Notebooks.'))

    # NB-specific argument: verbosity
    if 'nb' in name:
        group_verbose = parser.add_mutually_exclusive_group()
        help_verbose_1 = 'Set `verbose=1`: show all processes for entries with at least one used device.'
        help_verbose_2 = 'Set `verbose=2`: show all processes for all entries.'
        group_verbose.add_argument('-v', action='store_const', const=1, dest='verbose', help=help_verbose_1)
        group_verbose.add_argument('-V', action='store_const', const=2, dest='verbose', help=help_verbose_2)

    # Interval
    if 'watch' in name:
        help_interval = 'Interval (in seconds) between table updates.'
    else:
        help_interval = ('If provided, then the watch mode is used. '
                         'Value sets the interval (in seconds) between table updates.')
    parser.add_argument('-i', '--interval', '-n', '--watch', nargs='?', type=float, help=help_interval)

    help_window = 'Number of table updates to use for computing moving averages.'
    parser.add_argument('-w', '--window', nargs='?', type=int, help=help_window + linesep)

    # Show / hide columns by their aliases
    help_show = ('Additional columns to gather information about and show in the table.\n'
                 f'By default, following columns are not included: \n{formatter.excluded_names}')
    help_hide = ('Columns to exclude from the table, which also stops gathering information about them.\n'
                 f'By default, following columns are included: \n{formatter.included_names}')
    parser.add_argument('--show', nargs='*', help=help_show)
    parser.add_argument('--hide', nargs='*', help=help_hide)
    parser.add_argument('--show-all', action='store_const', const=True, help='Show all possible columns.')
    parser.add_argument('--hide-all', action='store_const', const=True,
                        help='Why would you ever want this? Does nothing.' + linesep)

    # Hidable columns
    help_changeable = 'Use this parameter to change this behavior.'
    help_hidable = f'By default, parts of rows with the same values as in previous row are hidden. {help_changeable}'
    parser.add_argument('--show-similar', action='store_const', const=False, dest='hide_similar', help=help_hidable)

    parser.add_argument('--hide-supheader', action='store_const', const=False, dest='add_supheader',
                        help=f'By default, we show current time, driver and CUDA versions. {help_changeable}')
    parser.add_argument('--hide-header', action='store_const', const=False, dest='add_header',
                        help=f'By default, we show a row with column names in the table. {help_changeable}')

    if 'watch' in name:
        parser.add_argument('--hide-footnote', action='store_const', const=False, dest='add_footnote',
                            help=f'By default, we show a row with total resource usage. {help_changeable}{linesep}')
    else:
        parser.add_argument('--show-footnote', action='store_const', const=True, dest='add_footnote',
                            help=f'Show a row with total system resource usage. {linesep}')

    parser.add_argument('--process-memory-format', type=str, default='GB',
                        help='Units of measurements for non-device memory stats, `GB` by default.')
    parser.add_argument('--device-memory-format', type=str,
                        help='Units of measurements for device memory stats, `MB` by default.' + linesep)

    # Separators
    group_separators = parser.add_mutually_exclusive_group()
    group_separators.add_argument('--show-separators', action='store_const', const=True, dest='separators',
                                  help='Turn on all the table separators.')
    group_separators.add_argument('--hide-separators', action='store_const', const=False, dest='separators',
                                  help='Turn off all the table separators.')

    parser.add_argument('--suppress-color', action='store_const', const=False, dest='force_styling',
                        help='Disable colors in the visualization.')

    # Merge defaults and passed arguments
    parser.set_defaults(**defaults)
    args = vars(parser.parse_args(argv))

    # Update
    separators = args.pop('separators')
    if separators is not None:
        for key in ['separate_supheader', 'separate_header', 'separate_index']:
            args[key] = separators

    if args.pop('hide_all'):
        pass

    # Update formatter with cmd arguments
    if args.pop('show_all'):
        formatter.include_all()

    for resource in args.pop('show'):
        if resource in formatter:
            formatter[resource] = True

    for resource in args.pop('hide'):
        if resource in formatter:
            formatter[resource] = False

    return formatter, args


if __name__ == '__main__':
    nbwatch()



""" Functions for running Jupyter Notebooks programmatically."""
#pylint: disable=import-outside-toplevel
import os
import time
from glob import glob
from textwrap import dedent

# Code fragments that are inserted in the notebook
CELL_INSERT_COMMENT = "# Cell inserted during automated execution"

# Connect to a shelve database for inputs/outputs providing
DB_CONNECT_CODE_CELL = """
    import os, shelve
    from dill import Pickler, Unpickler

    shelve.Pickler = Pickler
    shelve.Unpickler = Unpickler

    out_path_db = {}
"""
DB_CONNECT_CODE_CELL = dedent(DB_CONNECT_CODE_CELL)

# Insert inputs into the notebook
INPUTS_CODE_CELL = """
    # Inputs loading
    with shelve.open(out_path_db) as notebook_db:
        inputs = {**notebook_db}

        locals().update(inputs)
"""
INPUTS_CODE_CELL = dedent(INPUTS_CODE_CELL)

# Save notebook outputs
OUTPUTS_CODE_CELL = """
    # Output dict preparation
    output = {{}}
    outputs = {}

    for value_name in outputs:
        if value_name in locals():
            output[value_name] = locals()[value_name]

    with shelve.open(out_path_db) as notebook_db:
        notebook_db['outputs'] = output
"""
OUTPUTS_CODE_CELL = dedent(OUTPUTS_CODE_CELL)


def run_notebook(path, inputs=None, outputs=None, inputs_pos=1, working_dir = './', execute_kwargs=None,
                 out_path_db=None, out_path_ipynb=None, out_path_html=None, add_timestamp=True, hide_code_cells=False,
                 display_links=True, raise_exception=False, return_notebook=False):
    """ Execute a Jupyter Notebook programmatically.
    Heavily inspired by https://github.com/tritemio/nbrun.

    Intended to be an analog of `exec`, providing a way to inject / extract variables from the execution.
    Executed notebook is optionally saved to disk as `.ipynb`/`.html`: we strongly recommend always doing that.
    For a detailed description of how to do that, check the `inputs` and `outputs` parameters.

    Flag `raise_exception` defines behavior if the execution of the notebook is failed due to an exception.

    Under the hood, this function does the following:
        - Create an internal database to communicate variables (both `inputs` and `outputs`). Save `inputs` to it.
        - Add a cell with reading `inputs` from the database, add a cell for saving `outputs` to the database.
        - Execute notebook.
        - Handle exceptions.
        - Read `outputs` from the database.
        - Add a timestamp cell to the notebook, if needed.
        - Save the executed notebook as `.ipynb` and/or `.html`.
        - Return a dictionary with intermediate results, execution info and values of `outputs` variables.

    If there are no `inputs` nor `outputs`, a database is not created and additional cells are not inserted.
    If either of them is provided, then one of `out_path_ipynb` or `out_path_db` must be explicitly defined.

    Parameters
    ----------
    path : str
        Path to the notebook to execute.
    inputs : dict, optional
        Inputs for notebook execution: essentially, its `globals`.
        Must be a dictionary with variable names and their values; therefore, keys must be valid Python identifiers.
        Saved to a database, loaded in the notebook in a separate cell, that is inserted at `inputs_pos` position.
        Therefore, values must be serializable.
    outputs : str or iterable of str, optional
        List of notebook local variable names to return.
        Extracted from the notebook in a separate cell, that is inserted at the last position.
        If some of the variables don't exist, no errors are raised.
    inputs_pos : int, optional
        Position to insert the cell with `inputs` loading into the notebook.
    working_dir : str
        The working directory of starting the kernel.
    out_path_db : str, optional
        Path to save shelve database files without file extension.
        If not provided, then it is inferred from `out_path_ipynb`.
    out_path_ipynb : str, optional
        Path to save the output ipynb file.
    out_path_html : str, optional
        Path to save the output html file.
    execute_kwargs : dict, optional
        Parameters of `:class:ExecutePreprocessor`.
    add_timestamp : bool, optional
        Whether to add a cell with execution information at the beginning of the saved notebook.
    hide_code_cells : bool, optional
        Whether to hide the code cells in the saved notebook.
    display_links : bool, optional
        Whether to display links to the executed notebook and html at execution.
    raise_exception : bool, optional
        Whether to re-raise exceptions from the notebook.
    return_notebook : bool, optional
        Whether to return the notebook object from this function.

    Returns
    -------
    exec_res : dict
        Dictionary with the notebook execution results.
        It provides next information:
        - 'failed' : bool
           Whether the notebook execution failed.
        - 'outputs' : dict
           Saved notebook local variables.
           Is not presented in `exec_res` dict, if `outputs` argument is None.
        - 'failed cell number': int
           An error cell execution number (if notebook failed).
        - 'traceback' : str
           Traceback message from the notebook (if notebook failed).
        - 'notebook' : :class:`nbformat.notebooknode.NotebookNode`, optional
           Executed notebook object.
           Note that this output is provided only if `return_notebook` is True.
    """
    # pylint: disable=bare-except, lost-exception
    import nbformat
    from jupyter_client.manager import KernelManager
    from nbconvert.preprocessors import ExecutePreprocessor
    import shelve
    from dill import Pickler, Unpickler

    if inputs is not None or outputs is not None:
        # Set `out_path_db` value
        if out_path_db is None:
            if out_path_ipynb:
                out_path_db = os.path.splitext(out_path_ipynb)[0] + '_db'
            else:
                error_message = """\
                                Invalid value for `out_path_db` argument. If `inputs` or `outputs` are provided,
                                then you need to provide `out_path_db` or `out_path_ipynb` arguments."""
                error_message = dedent(error_message)
                raise ValueError(error_message)

        # Create a shelve database
        shelve.Pickler = Pickler
        shelve.Unpickler = Unpickler

        with shelve.open(out_path_db) as notebook_db:
            notebook_db.clear()

    if isinstance(outputs, str):
        outputs = [outputs]

    execute_kwargs = execute_kwargs or {'timeout': -1}
    executor = ExecutePreprocessor(**execute_kwargs)
    kernel_manager = KernelManager()

    # Notebook preparation:
    # Read the notebook, insert a cell with inputs, insert another cell for outputs extraction
    notebook = nbformat.read(path, as_version=4)

    if hide_code_cells:
        notebook["metadata"].update({"hide_input": True})

    if inputs is not None:
        # Save `inputs` in the shelve database and create a cell in the notebook
        # for parameters extraction
        with shelve.open(out_path_db) as notebook_db:
            notebook_db.update(inputs)

        code = CELL_INSERT_COMMENT + DB_CONNECT_CODE_CELL.format(repr(out_path_db)) + INPUTS_CODE_CELL
        notebook['cells'].insert(inputs_pos, nbformat.v4.new_code_cell(code))

    if outputs is not None:
        # Create a cell to extract outputs from the notebook
        # It saves locals from the notebook with preferred names in the shelve database
        # This cell will be executed in error case too
        code = CELL_INSERT_COMMENT + \
               (DB_CONNECT_CODE_CELL.format(repr(out_path_db)) if not inputs else "") + \
               OUTPUTS_CODE_CELL.format(outputs)
        output_cell = nbformat.v4.new_code_cell(code)
        notebook['cells'].append(output_cell)

    # Execute the notebook
    start_time = time.time()
    exec_failed = False
    try:
        executor.preprocess(notebook, {'metadata': {'path': working_dir}}, km=kernel_manager)
    except:
        exec_failed = True

        # Save notebook outputs in the shelve db
        if outputs is not None:
            executor.kc = kernel_manager.client() # For compatibility with 5.x.x version of `nbconvert`
            executor.preprocess_cell(output_cell, {'metadata': {'path': working_dir}}, -1)

        if raise_exception:
            raise
    finally:
        kernel_manager.shutdown_kernel()

        # Extract information from the database and remove it (if exists)
        if outputs is not None:
            with shelve.open(out_path_db) as notebook_db:
                outputs_values = notebook_db.get('outputs', {})

        if out_path_db is not None:
            db_paths = glob(out_path_db + '*')

            for path_ in db_paths:
                os.remove(path_)

        # Check if something went wrong
        failed, error_cell_num, traceback_message = extract_traceback(notebook=notebook)

        if exec_failed:
            failed = True
            traceback_message += '\nNotebook execution failed\n'

        # Prepare execution results: execution state, notebook outputs and error info (if exists)
        if failed:
            exec_res = {'failed': failed, 'failed cell number': error_cell_num, 'traceback': traceback_message}
        else:
            exec_res = {'failed': failed, 'failed cell number': None, 'traceback': ''}

        if outputs is not None:
            exec_res['outputs'] = outputs_values

        if add_timestamp:
            timestamp = (f"**Executed:** {time.ctime(start_time)}<br>"
                         f"**Duration:** {time.strftime('%H:%M:%S', time.gmtime(time.time() - start_time))}<br>"
                         f"**Autogenerated from:** [{path}]\n\n---")
            timestamp_cell = nbformat.v4.new_markdown_cell(timestamp)
            notebook['cells'].insert(0, timestamp_cell)

        # Save the executed notebook/HTML to disk
        if out_path_ipynb is not None:
            save_notebook(notebook=notebook, out_path_ipynb=out_path_ipynb, display_link=display_links)
        if out_path_html is not None:
            notebook_to_html(notebook=notebook, out_path_html=out_path_html, display_link=display_links)

        if return_notebook:
            exec_res['notebook'] = notebook
        return exec_res

# Save notebook functions
def save_notebook(notebook, out_path_ipynb, display_link):
    """ Save an instance of :class:`nbformat.notebooknode.NotebookNode` as ipynb file."""
    import nbformat
    from IPython.display import display, FileLink

    with open(out_path_ipynb, 'w', encoding='utf-8') as file:
        nbformat.write(notebook, file)

    if display_link:
        display(FileLink(out_path_ipynb))

def notebook_to_html(notebook, out_path_html, display_link):
    """ Save an instance of :class:`nbformat.notebooknode.NotebookNode` as html file."""
    from nbconvert import HTMLExporter
    from IPython.display import display, FileLink

    html_exporter = HTMLExporter()
    body, _ = html_exporter.from_notebook_node(notebook)

    with open(out_path_html, 'w', encoding='utf-8') as f:
        f.write(body)

    if display_link:
        display(FileLink(out_path_html))


def extract_traceback(notebook):
    """ Extracts information about an error from the notebook.

    Parameters
    ----------
    notebook: :class:`nbformat.notebooknode.NotebookNode`
        Executed notebook to find an error traceback.

    Returns
    -------
    bool
        Whether the executed notebook has an error traceback.
    int or None
        Number of a cell with a traceback.
        If None, then the notebook doesn't contain an error traceback.
    str
        Error traceback if exists.
    """
    for cell in notebook['cells']:
        # Find a cell output with a traceback and extract the traceback
        outputs = cell.get('outputs', [])

        for output in outputs:
            traceback = output.get('traceback', [])

            if traceback:
                traceback = '\n'.join(traceback)
                return True, cell['execution_count'], traceback

    return False, None, ""



"""
NBTools is a collection of tools for using in notebooks and inspect running notebooks.
"""
import re
from setuptools import setup, find_packages


with open('nbtools/__init__.py', 'r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='py-nbtools',
    packages=find_packages(),
    version=version,
    url='https://github.com/analysiscenter/nbtools',
    license='Apache License 2.0',
    author='Sergey Tsimfer',
    author_email='sergeytsimfer@gmail.com',
    description='A collection of tools for using inside Jupyter Notebooks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    platforms='any',
    install_requires=[
        'nvidia-ml-py3>=7.352',
        'blessed>=1.17',
        'psutil>=5.6',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: System :: Monitoring',
    ],
    entry_points={
        'console_scripts': [
            'nbstat = nbtools.nbstat.cli:nbstat',
            'nbwatch = nbtools.nbstat.cli:nbwatch',
            'devicestat = nbtools.nbstat.cli:devicestat',
            'devicewatch = nbtools.nbstat.cli:devicewatch',
        ],
    },
)



""" A class to define resources to fetch from the system, as well as the table structure. """
from copy import deepcopy
from operator import attrgetter

from .resource import Resource



class ResourceFormatter(list):
    """ An ordered sequence to define which resources to request from the system, as well as
    how to structure them into formatted table.

    Each element is a dictionary with mandatory `resource` and `include` keys, and defines whether this
    `resource` must be fetched/included in the table based on the value of `include` flag.
    Other keys are used to format this column: refer to `ResourceTable.format` method for more about that.

    Note that we can't use `dict` (with resource as key) for this class directly as it would prohibit repeating columns.
    Despite that, we overload `getitem` to provide a dict-like interface for checking whether the `resource` should be
    requested from the system. For example, `formatter[Resource.DEVICE_TEMP]` returns True if it should be fetched.

    We also overload `setitem` to change value of `include` flag for a given resource. That provides simple API for
    modifying existing formatters.
    For example, `formatter[Resource.DEVICE_TEMP] = True` turns on the device temperature column, if it is
    present in the `formatter`.

    Note that it does not change the order of elements: that is the key feature and allows us to define
    'templates` of table formatters, where some of the elements have `include` flag set to False by default.
    Changing the value of `include` flag adds required resources in the table in pre-defined positions.
    Refer to `NBSTAT_FORMATTER` for example.
    """
    def __getitem__(self, key):
        key = Resource.parse_alias(key)

        if isinstance(key, Resource):
            for entry in self:
                if entry['resource'] is key:
                    if entry['include']:
                        return True
            return False

        result = super().__getitem__(key)
        return ResourceFormatter(result) if isinstance(result, list) else result

    def get(self, key, default=None):
        """ `getitem` with `default`. """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        """ Overloaded `in` operator. """
        try:
            _ = self[key]
            return True
        except KeyError:
            return False

    def __setitem__(self, key, value):
        """ If `key` is a resource (or an alias), set the values of `include` flag for this resource. """
        key = Resource.parse_alias(key)

        if isinstance(key, str):
            raise KeyError(f'Key `{key}` is not recognized!')

        if isinstance(key, Resource):
            for entry in self:
                if entry['resource'] is key:
                    entry['include'] = value
                    return None
            raise KeyError(f'Entry `{key}` is not in the formatter!')
        return super().__setitem__(key, value)

    def update(self, other=None, **kwargs):
        """ Change multiple values at once. """
        other = other if other is not None else {}
        kwargs.update(other)

        for key, value in kwargs.items():
            self[key] = value

        if isinstance(other, ResourceFormatter):
            self.extend(other)

    def copy(self):
        """ Deep copy of the formatter. Used to not mess up the original formatter. """
        return deepcopy(self)

    def include_all(self):
        """ Turn on collection of all present resources. """
        for column in self:
            column['include'] = True

    @property
    def included_only(self):
        """ Return formatter with elements which `include` flag is set to True.
        Also removes subsequent duplicates of table delimiters.
        """
        formatter = []
        for column in self:
            included = column['include']
            if not included:
                continue

            resource = column['resource']

            if 'TABLE_DELIMITER' in resource.name and len(formatter) > 0:
                previous_resource = formatter[-1]['resource']
                if 'TABLE_DELIMITER' in previous_resource.name:
                    formatter[-1]['resource'] = max(resource, previous_resource, key=attrgetter('value'))
                    continue
            formatter.append(column)

        if 'TABLE_DELIMITER' in formatter[-1]['resource'].name:
            formatter.pop()
        return formatter

    @property
    def names(self):
        """ Aliases of all resources in `self`. """
        return [Resource.RESOURCE_TO_ALIAS[item['resource']] for item in self]

    @property
    def included_names(self):
        """ Aliases of included resources in `self`. """
        return [Resource.RESOURCE_TO_ALIAS[item['resource']] for item in self.included_only
                if 'TABLE_DELIMITER' not in item['resource'].name]

    @property
    def excluded_names(self):
        """ Aliases of not included resources in `self`. """
        return [Resource.RESOURCE_TO_ALIAS[item['resource']] for item in self
                if item['include'] is False and 'TABLE_DELIMITER' not in item['resource'].name]

    def toggle_bars(self):
        """ Change resources to use `bar` representation. """
        for item in self:
            if 'bar' in item:
                item['bar'] = not item['bar']


NBSTAT_FORMATTER = ResourceFormatter([
    # Notebook/script name
    {'resource' : Resource.NAME, 'include' : True, 'hidable': True},
    {'resource' : Resource.PATH, 'include' : False, 'hidable': True},
    {'resource' : Resource.CMDLINE, 'include' : False, 'hidable': True},

    # Process info
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.TYPE, 'include' : True},
    {'resource' : Resource.PID, 'include' : True},
    {'resource' : Resource.PPID, 'include' : False},
    {'resource' : Resource.NGID, 'include' : False},
    {'resource' : Resource.PYTHON_PPID, 'include' : False},
    {'resource' : Resource.HOST_PID, 'include' : False},
    {'resource' : Resource.KERNEL, 'include' : False},
    {'resource' : Resource.STATUS, 'include' : False, 'min_width' : 10},

    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.CREATE_TIME, 'include' : False},

    # Process resource usage
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.CPU, 'include' : False, 'min_width' : 5},
    {'resource' : Resource.RSS, 'include' : True, 'min_width' : 8},

    # Process device usage
    {'resource' : Resource.TABLE_DELIMITER2, 'include' : True},
    {'resource' : Resource.DEVICE_SHORT_ID, 'include' : True},
    {'resource' : Resource.DEVICE_PROCESS_PID, 'include' : False},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_PROCESS_MEMORY_USED, 'include' : True},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_UTIL, 'include' : True, 'min_width' : 5, 'bar': False},
    {'resource' : Resource.DEVICE_UTIL_MA, 'include' : False, 'min_width' : 5, 'bar': False},
    {'resource' : Resource.DEVICE_TEMP, 'include' : True, 'min_width' : 5},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_POWER_USED, 'include' : False},
    {'resource' : Resource.DEVICE_FAN, 'include' : False, 'min_width' : 4},
])


DEVICESTAT_FORMATTER = ResourceFormatter([
    # Device usage
    {'resource' : Resource.DEVICE_ID, 'include' : True, 'hidable' : True},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_UTIL, 'include' : True, 'hidable' : True, 'min_width' : 5, 'bar': False},
    {'resource' : Resource.DEVICE_UTIL_MA, 'include' : False, 'min_width' : 5, 'bar': False},
    {'resource' : Resource.DEVICE_TEMP, 'include' : True, 'hidable' : True, 'min_width' : 5},
    {'resource' : Resource.DEVICE_POWER_USED, 'include' : False, 'hidable' : True},
    {'resource' : Resource.DEVICE_FAN, 'include' : False, 'hidable' : True, 'min_width' : 4},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},

    # Individual processes for each device
    {'resource' : Resource.DEVICE_PROCESS_MEMORY_USED, 'include' : True},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},

    # Notebook/script name
    {'resource' : Resource.NAME, 'include' : True},
    {'resource' : Resource.CMDLINE, 'include' : False, 'hidable': True},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},

    # Process info
    {'resource' : Resource.TYPE, 'include' : False},
    {'resource' : Resource.PID, 'include' : False},
    {'resource' : Resource.PPID, 'include' : False},
    {'resource' : Resource.NGID, 'include' : False},
    {'resource' : Resource.PYTHON_PPID, 'include' : False},
    {'resource' : Resource.DEVICE_PROCESS_PID, 'include' : True},
    {'resource' : Resource.HOST_PID, 'include' : False},
    {'resource' : Resource.KERNEL, 'include' : False},
    {'resource' : Resource.STATUS, 'include' : False},
    {'resource' : Resource.CREATE_TIME, 'include' : False},

    # Process resource usage
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.CPU, 'include' : False, 'min_width' : 5},
    {'resource' : Resource.RSS, 'include' : False, 'min_width' : 10},

])


GPUSTAT_FORMATTER = ResourceFormatter([
    # Device usage
    {'resource' : Resource.DEVICE_ID, 'include' : True},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_UTIL, 'include' : True, 'min_width': 5, 'bar': False},
    {'resource' : Resource.DEVICE_TEMP, 'include' : True, 'min_width': 5},
    {'resource' : Resource.DEVICE_POWER_USED, 'include' : False},
    {'resource' : Resource.TABLE_DELIMITER1, 'include' : True},
    {'resource' : Resource.DEVICE_MEMORY_USED, 'include' : True},
])
